@@ -1,6 +1,7 @@
 use v6;
+use fatal;
 
-unit class Getopt::Long:ver<0.1.0>;
+unit class Getopt::Long:ver<0.1.1>;
 
 my sub null-converter(Str:D $value --> Str) {
 	return $value;
@@ -10,41 +11,6 @@ my sub int-converter(Str:D $value --> Int) {
 	return $value.Int;
 }
 
-my sub extended-int-converter(Str:D $value --> Int) {
-	grammar Extended {
-		token TOP {
-			<sign> <bulk>
-			{ make $<sign>.ast * $<bulk>.ast; }
-		}
-
-		token sign {
-			$<char>=<[+-]>?
-			{ make ($<char> eq '-' ?? -1 !! 1) }
-		}
-		token bulk {
-			[ <hex> | <octal> | <binary> || <fallback> ]
-			{ make $/.values[0].ast }
-		}
-		token hex {
-			:i '0x' $<num>=[<[0..9A..F]>+]
-			{ make :16(~$<num>) }
-		}
-		token octal {
-			'0' $<num>=[<[0..7]>+]
-			{ make :8(~$<num>) }
-		}
-		token binary {
-			:i '0b' $<num>=[<[01]>+]
-			{ make :2(~$<num>) }
-		}
-		token fallback {
-			\d+
-			{ make $/.Str.Int }
-		}
-	}
-	return Extended.parse($value).ast // Int;
-}
-
 my sub rat-converter(Str:D $value --> Rat) {
 	return $value.Rat;
 }
@@ -53,8 +19,8 @@ my sub num-converter(Str:D $value --> Num) {
 	return $value.Num;
 }
 
-my sub rat-num-converter(Str:D $value --> Real) {
-	return val($value) ~~ Rat|Int ?? $value.Rat !! $value.Num;
+my sub real-converter(Str:D $value --> Real) {
+	return $value.Real;
 }
 
 my sub maybe-converter(Str:D $value --> Any) {
@@ -85,7 +51,7 @@ my class CountStore does Store {
 my class ArrayStore does Store {
 	has Any:U $.type = $!converter.returns;
 	method store-direct(Any:D $value, Hash:D $hash) {
-		$hash{$!key} //= Array[$!type].new;
+		$hash{$!key} //= $!type === Any ?? Array !! Array[$!type].new;
 		$hash{$!key}.push($value);
 	}
 }
@@ -94,12 +60,12 @@ my class HashStore does Store {
 	has Any:U $.type = $!converter.returns;
 	method store-convert(Any:D $pair, Hash:D $hash) {
 		my ($key, $value) = $pair.split('=', 2);
-		$hash{$!key} //= Hash[$!type].new;
+		$hash{$!key} //= $!type === Any ?? Hash !! Hash[$!type].new;
 		$hash{$!key}{$key} = $!converter($value);
 	}
 	method store-direct(Any:D $pair, Hash:D $hash) {
 		my ($key, $value) = $pair.split('=', 2);
-		$hash{$!key} //= Hash[$!type].new;
+		$hash{$!key} //= $!type === Any ?? Hash !! Hash[$!type].new;
 		$hash{$!key}{$key} = $value;
 	}
 }
@@ -121,6 +87,10 @@ has Option:D %!options;
 
 submethod BUILD(:%!options) { }
 
+method !options {
+	return %!options;
+}
+
 my %store-for = (
 	'%' => HashStore,
 	'@' => ArrayStore,
@@ -172,9 +142,8 @@ my grammar Argument {
 	my %converter-for-format = (
 		i => &int-converter,
 		s => &null-converter,
-		f => &rat-num-converter,
+		f => &real-converter,
 		r => &rat-converter,
-		o => &extended-int-converter,
 	);
 
 	token type {
@@ -228,11 +197,12 @@ method new-from-patterns(@patterns, *%args) {
 }
 
 my %converter-for-type{Any:U} = (
-	(Int) => &int-converter,
-	(Rat) => &rat-converter,
-	(Num) => &num-converter,
-	(Str) => &null-converter,
-	(Any) => &maybe-converter,
+	(Int)  => &int-converter,
+	(Rat)  => &rat-converter,
+	(Num)  => &num-converter,
+	(Real) => &real-converter,
+	(Str)  => &null-converter,
+	(Any)  => &maybe-converter,
 );
 
 my role Formatted {
@@ -248,7 +218,7 @@ my sub parse-parameter(Parameter $param) {
 	if $param ~~ Formatted {
 		my $pattern = $param.format;
 		if Argument.parse($pattern, :rule('argument')) -> $match {
-			return @($match.ast).map(&make-option.assuming(@names)).flat;
+			return make-option(@names, |@($match.ast))
 		}
 		else {
 			die "Couldn't parse '$pattern'";
@@ -289,16 +259,32 @@ my sub parse-parameter(Parameter $param) {
 	}
 }
 
+my role Parsed {
+	has Getopt::Long:D $.getopt is required;
+}
+
+multi sub trait_mod:<is>(Sub $sub, :$getopt!) is export(:DEFAULT, :traits) {
+	$sub does Parsed(Getopt::Long.new-from-sub($sub));
+}
+
 method new-from-sub(Sub $main) {
+	return $main.getopt if $main ~~ Parsed;
+
+	my sub get-options($candidate) {
+		if $candidate ~~ Parsed {
+			return $candidate.getopt!options.values;
+		}
+		else {
+			return $candidate.signature.params.grep(*.named).map(&parse-parameter).flat;
+		}
+	}
 	my %options;
 	for $main.candidates -> $candidate {
-		for $candidate.signature.params.grep(*.named) -> $param {
-			for parse-parameter($param) -> $option {
-				if %options{$option.name}:exists and %options{$option.name} !eqv $option {
-					die "Can't merge arguments for {$option.name}";
-				}
-				%options{$option.name} = $option;
+		for get-options($candidate) -> $option {
+			if %options{$option.name}:exists and %options{$option.name} !eqv $option {
+				die "Can't merge arguments for {$option.name}";
 			}
+			%options{$option.name} = $option;
 		}
 	}
 	return self.new(:%options);
@@ -314,16 +300,33 @@ method get-options(@args is copy, :%hash, :named-anywhere(:$permute) = False, :$
 		sub take-value($option, $value) {
 			$option.store($value, %hash);
 			$consumed++;
+
+			CATCH {
+				when X::Str::Numeric {
+					$_ does role {
+						method message() {
+							"Cannot convert --{$option.name} argument to number: $.reason $.source-indicator";
+						}
+					}
+					.rethrow;
+				}
+				when X::Numeric::CannotConvert {
+					$_ does role {
+						method message() {
+							"Cannot convert --{$option.name} argument $.source to {$.target // $.target.perl}: $.reason"
+						}
+					}
+					.rethrow;
+				}
+			}
 		}
 		sub take-args($option) {
 			while @args && $consumed < $option.arity.min {
-				$option.store(@args.shift, %hash);
-				$consumed++;
+				take-value($option, @args.shift);
 			}
 
 			while @args && $consumed < $option.arity.max && !@args[0].starts-with('--') {
-				$option.store(@args.shift, %hash);
-				$consumed++;
+				take-value($option, @args.shift);
 			}
 
 			if $consumed == 0 && $option.arity.min == 0 {
@@ -405,6 +408,9 @@ our sub get-options-from(@args, *@elements, :$overwrite, *%config) is export(:DE
 			}
 			@options.push: $key;
 		}
+		default {
+			die "Unknown element type: " ~ $element.perl;
+		}
 	}
 	my $getopt = Getopt::Long.new-from-patterns(@options);
 	return $getopt.get-options(@args, |%config, :%hash, :write-args($overwrite ?? @args !! Any));
@@ -807,18 +813,6 @@ sequence of digits.
 
 =end item2
 
-=begin item2
-o
-
-Extended integer, Perl style. This can be either an optional leading
-plus or minus sign, followed by a sequence of digits, or an octal
-string (a zero, optionally followed by '0', '1', .. '7'), or a
-hexadecimal string (C<0x> followed by '0' .. '9', 'a' .. 'f', case
-insensitive), or a binary string (C<0b> followed by a series of '0'
-and '1').
-
-=end item2
-
 =begin item2
 r
 
